/*	XGBroadcast.cpp
 *
 *		This is a global class which uses static (global) methods for
 *	broadcasting messages
 */

/**********************************************************************

  YAAF is Copyright 1997-1999 by William Woody.

  Redistribution and use in source and binary forms, with or without 
  modification, are permitted provided that the following conditions 
  are met:

  Redistributions of source code must retain the above copyright notice, 
  this list of conditions and the following disclaimer.

  Neither the name of the copyright holder or other contributors may
  be used to promote products derived from this software without
  specific written prior permission.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES 
  OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND 
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS 
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN 
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN 
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE 
  SOFTWARE.
  
  Contact:   woody@alumni.caltech.edu
  Web page:  http://www.pandawave.com/yaaf/

**********************************************************************/

#include <XConfig.h>			// must be first to optimize compile on MSVC++
#include <XEvent.h>

/************************************************************************/
/*																		*/
/*	Globals																*/
/*																		*/
/************************************************************************/

XGSend XGBroadcast::gSend;
XGSDynArray<_PostRecord> XGBroadcast::gPostStack;

/************************************************************************/
/*																		*/
/*	PostBroadcast Support												*/
/*																		*/
/************************************************************************/

/*$	XGBroadcast::PostBroadcast
 *
 *		This takes the arguments and creates a _PostRecord which
 *	contains the arguments, for later posting.
 *
 *		WARNING: As the messages are sent later, putting the 'parg'
 *	object on the stack would be a problem.
 */

void XGBroadcast::PostBroadcast(long msg, long arg, void *parg)
{
	_PostRecord p;
	
	p.msg = msg;
	p.arg = arg;
	p.parg = parg;
	gPostStack += p;
}

/*$	XGBroadcast::SendPostBroadcast
 *
 *		Run the list and send all the messages that were delayed. This
 *	is automatically called from the internal YAAF event loop, so it's
 *	not necessary for you to call this operation yourself.
 */

void XGBroadcast::SendPostBroadcast()
{
	long i,len;
	
	len = gPostStack.Length();
	for (i = 0; i < len; i++) {
		gSend.SendMessage(gPostStack[i].msg,
						  gPostStack[i].arg,
						  gPostStack[i].parg);
	}
	gPostStack.DeleteAll();
}
