/*	XGDocForm.cpp
 *
 *		This encapsulates the document handling functionality of a window
 *	by defining the basic framework inside of a form.
 */

/**********************************************************************

  YAAF is Copyright 1997-1999 by William Woody.

  Redistribution and use in source and binary forms, with or without 
  modification, are permitted provided that the following conditions 
  are met:

  Redistributions of source code must retain the above copyright notice, 
  this list of conditions and the following disclaimer.

  Neither the name of the copyright holder or other contributors may
  be used to promote products derived from this software without
  specific written prior permission.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES 
  OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND 
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS 
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN 
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN 
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE 
  SOFTWARE.
  
  Contact:   woody@alumni.caltech.edu
  Web page:  http://www.pandawave.com/yaaf/

**********************************************************************/

#include <XConfig.h>
#include <XApplication.h>
#include <XAppMenu.h>
#include <XForm.h>
#include <XError.h>
#include <XDialog.h>
#include <XDocument.h>
#include <XMessages.h>

/************************************************************************/
/*																		*/
/*	Window Save Dialog Specification									*/
/*																		*/
/************************************************************************/

/*	GSaveDialog
 *
 *		The data which specifies my 'save' dailog box
 */

static unsigned char GSaveDialog[] = {
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x64, /* 00000000: .......d */
    0x00, 0x00, 0x00, 0x54, 0x64, 0x6C, 0x6F, 0x67, /* 00000008: ...Tdlog */
    0x00, 0x00, 0x00, 0x12, 0x00, 0x00, 0x00, 0x00, /* 00000010: ........ */
    0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 00000018: ........ */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 00000020: ........ */
    0x00, 0x00, 0x00, 0x00, 0x01, 0x5E, 0x00, 0x00, /* 00000028: .....^.. */
    0x00, 0x76, 0x00, 0x00, 0x01, 0x5E, 0x00, 0x00, /* 00000030: .v...^.. */
    0x00, 0x76, 0x00, 0x00, 0x01, 0x5E, 0x00, 0x00, /* 00000038: .v...^.. */
    0x00, 0x76, 0x00, 0x00, 0x01, 0x5E, 0x00, 0x00, /* 00000040: .v...^.. */
    0x00, 0x76, 0x0C, 0x53, 0x61, 0x76, 0x65, 0x20, /* 00000048: .v.Save. */
    0x57, 0x69, 0x6E, 0x64, 0x6F, 0x77, 0x3F, 0x00, /* 00000050: Window?. */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x64, /* 00000058: .......d */
    0x00, 0x00, 0x00, 0x65, 0x00, 0x00, 0x00, 0x42, /* 00000060: ...e...B */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x32, /* 00000068: .......2 */
    0x69, 0x63, 0x6F, 0x6E, 0x00, 0x00, 0x00, 0x00, /* 00000070: icon.... */
    0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, /* 00000078: ........ */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x18, /* 00000080: ........ */
    0x00, 0x00, 0x00, 0x10, 0x00, 0x00, 0x00, 0x38, /* 00000088: .......8 */
    0x00, 0x00, 0x00, 0x30, 0xFF, 0xFF, 0xFF, 0xFF, /* 00000090: ...0.... */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x02, 0x00, 0x00, /* 00000098: ........ */
    0x00, 0x20, 0x00, 0x00, 0x00, 0x20, 0x00, 0x00, /* 000000A0: ........ */
    0x00, 0x7A, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 000000A8: .z...... */
    0x00, 0x6A, 0x74, 0x65, 0x78, 0x74, 0x00, 0x00, /* 000000B0: .jtext.. */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, /* 000000B8: ........ */
    0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, /* 000000C0: ........ */
    0x00, 0x50, 0x00, 0x00, 0x00, 0x10, 0xFF, 0xFF, /* 000000C8: .P...... */
    0xFF, 0xF2, 0x00, 0x00, 0x00, 0x50, 0xFF, 0xFF, /* 000000D0: .....P.. */
    0xFF, 0xFF, 0x00, 0x00, 0x3F, 0x53, 0x68, 0x6F, /* 000000D8: ....?Sho */
    0x75, 0x6C, 0x64, 0x20, 0x74, 0x68, 0x65, 0x20, /* 000000E0: uld.the. */
    0x63, 0x6F, 0x6E, 0x74, 0x65, 0x6E, 0x74, 0x73, /* 000000E8: contents */
    0x20, 0x6F, 0x66, 0x20, 0x5E, 0x30, 0x20, 0x62, /* 000000F0: .of.^0.b */
    0x65, 0x20, 0x73, 0x61, 0x76, 0x65, 0x64, 0x20, /* 000000F8: e.saved. */
    0x62, 0x65, 0x66, 0x6F, 0x72, 0x65, 0x20, 0x74, /* 00000100: before.t */
    0x68, 0x65, 0x20, 0x77, 0x69, 0x6E, 0x64, 0x6F, /* 00000108: he.windo */
    0x77, 0x20, 0x69, 0x73, 0x20, 0x63, 0x6C, 0x6F, /* 00000110: w.is.clo */
    0x73, 0x65, 0x64, 0x3F, 0x00, 0x00, 0x00, 0x00, /* 00000118: sed?.... */
    0x00, 0x00, 0x00, 0x44, 0x00, 0x00, 0x00, 0x00, /* 00000120: ...D.... */
    0x00, 0x00, 0x00, 0x34, 0x70, 0x62, 0x74, 0x6E, /* 00000128: ...4pbtn */
    0x00, 0x00, 0x00, 0x64, 0x00, 0x00, 0x00, 0x00, /* 00000130: ...d.... */
    0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, /* 00000138: ........ */
    0x00, 0x00, 0x01, 0x10, 0x00, 0x00, 0x00, 0x56, /* 00000140: .......V */
    0x00, 0x00, 0x01, 0x54, 0x00, 0x00, 0x00, 0x72, /* 00000148: ...T...r */
    0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, /* 00000150: ........ */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x53, /* 00000158: .......S */
    0x61, 0x76, 0x65, 0x00, 0x00, 0x00, 0x00, 0x46, /* 00000160: ave....F */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x36, /* 00000168: .......6 */
    0x70, 0x62, 0x74, 0x6E, 0x00, 0x00, 0x00, 0x65, /* 00000170: pbtn...e */
    0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, /* 00000178: ........ */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xC4, /* 00000180: ........ */
    0x00, 0x00, 0x00, 0x56, 0x00, 0x00, 0x01, 0x08, /* 00000188: ...V.... */
    0x00, 0x00, 0x00, 0x72, 0xFF, 0xFF, 0xFF, 0xFF, /* 00000190: ...r.... */
    0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 00000198: ........ */
    0x00, 0x00, 0x06, 0x43, 0x61, 0x6E, 0x63, 0x65, /* 000001A0: ...Cance */
    0x6C, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 000001A8: l....... */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x3A, 0x70, 0x62, /* 000001B0: .....:pb */
    0x74, 0x6E, 0x00, 0x00, 0x00, 0x66, 0x00, 0x00, /* 000001B8: tn...f.. */
    0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, /* 000001C0: ........ */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x28, 0x00, 0x00, /* 000001C8: .....(.. */
    0x00, 0x56, 0x00, 0x00, 0x00, 0x8C, 0x00, 0x00, /* 000001D0: .V...... */
    0x00, 0x72, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, /* 000001D8: .r...... */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 000001E0: ........ */
    0x0A, 0x44, 0x6F, 0x6E, 0x27, 0x74, 0x20, 0x53, /* 000001E8: .Don't.S */
    0x61, 0x76, 0x65, 0x00                          /* 000001F0: ave.     */
}; /* _R56526573_0083 */

/************************************************************************/
/*																		*/
/*	Construction/Destruction											*/
/*																		*/
/************************************************************************/

/*$	XGDocForm::XGDocForm
 *
 *		Document form construction. The #secondary# parameter is used to
 *	indicate if the form associated with this window is a "secondary"
 *	form--that is, if this window is not a primary document editor.
 *
 *		If this is a primary document editor window, then when the last
 *	one of these are closed, this asks the user if the document should be
 *	saved if it's dirty. If this is a secondary editor, though, closing
 *	this window doesn't affect anything, and when the last primary window
 *	is closed, all the secondary windows are automatically closed as well.
 */

XGDocForm::XGDocForm(bool secondary)
{
	fSecondary = secondary;
	fDocument = NULL;
	fNext = NULL;
}

/*$	XGDocForm::~XGDocForm
 *
 *		Time to go away.
 */

XGDocForm::~XGDocForm()
{
	DetachDocument();
}

/************************************************************************/
/*																		*/
/*	Editor Event Support												*/
/*																		*/
/************************************************************************/

/*$	XGDocForm::AttachDocument
 *
 *		Attach the specified document to me.
 */

void XGDocForm::AttachDocument(XGDocument *doc)
{
	Assert(fDocument == NULL);
	
	fDocument = doc;
	fNext = doc->fForm;
	doc->fForm = this;
}

/*$	XGDocForm::DetachDocument
 *
 *		Detach this document. If this is the last primary form, this deletes
 *	the document, which deletes all the attached forms. This is nromally
 *	called internally by YAAF.
 */

void XGDocForm::DetachDocument()
{
	XGDocForm *foo;
	
	Assert(fDocument != NULL);			// make sure things are called right
	
	if (fDocument->fForm == this) {		// remove me from the document list
		fDocument->fForm = fNext;
	} else {
		foo = fDocument->fForm;
		while (foo != NULL) {
			if (foo->fNext == this) {
				foo->fNext = fNext;
				break;
			}
			foo = foo->fNext;
		}
	}
	
	/*
	 *	Now that this form is detached, determine if there are any
	 *	primary form objects associated with this document. If not,
	 *	delete the document
	 */
	
	for (foo = fDocument->fForm; foo != NULL; foo = foo->fNext) {
		if (!(foo->fSecondary)) break;
	}
	if (foo == NULL) delete fDocument;

	fDocument = NULL;					// and clear the document flag
}

/************************************************************************/
/*																		*/
/*	Window close management												*/
/*																		*/
/************************************************************************/

/*$	XGDocForm::DoQueryDestroyWindow
 *
 *		This is called by my window before it's closed. In this implementation,
 *	I get the name of the document, determine if the document needs to be
 *	saved, and ask the user to save it. This is called internally by YAAF.
 */

bool XGDocForm::DoQueryDestroyWindow()
{
	char buffer[256];
	long val;
	long count;
	XGDocForm *form;
	
	if (fDocument && (!fSecondary)) {
		form = fDocument->fForm;
		count = 0;
		while (form) {
			if (!(form->fSecondary)) count++;
			form = form->fNext;
		}
		
		if ((count == 1) && (fDocument->IsDocumentDirty())) {
			/*
			 *	There is just this document editor. Ask the user
			 *	to save the window
			 */
			
			GetWindow()->GetTitle(buffer);
			val = XGDialog::ModalAlert(GSaveDialog,buffer,NULL,NULL,NULL);
			if (val == 102) {
				/*
				 *	Don't save file
				 */
				
				return true;
			}
			if (val == 101) {
				/*
				 *	Cancel close operation
				 */
				
				return false;
			}
			
			/*
			 *	Save file first.
			 */
			
			return fDocument->DoDocumentSave();
		}
	}
	return true;
}

/*$	XGDocForm::GetParentDispatch
 *
 *		Redirect messages upwards again. This dispatches messages upwards
 *	to the document if one exists.
 */

XGDispatch *XGDocForm::GetParentDispatch()
{
	if (fDocument) return fDocument;
	else return XGAppCore::GetApplication();
}

/************************************************************************/
/*																		*/
/*	Events																*/
/*																		*/
/************************************************************************/

/*$	XGDocForm::DoDocumentChanged
 *
 *		Document has changed
 */

void XGDocForm::DoDocumentChanged()
{
}

/*$	XGDocForm::DoDocumentRevert
 *
 *		Document has reverted
 */

void XGDocForm::DoDocumentRevert()
{
}

/*$	XGDocForm::DoDocumentRenamed
 *
 *		Document has been renamed
 */

void XGDocForm::DoDocumentRenamed()
{
}

/*$	XGDocForm::DoDocumentSaved
 *
 *		Document has been saved
 */

void XGDocForm::DoDocumentSaved()
{
}

BEGINDISPATCH(XGDocForm,XGForm)
ENDDISPATCH

