/*	XGToolTip.cpp
 *
 *		This handles the tool tip object.
 */

/**********************************************************************

  YAAF is Copyright 1997-1999 by William Woody.

  Redistribution and use in source and binary forms, with or without 
  modification, are permitted provided that the following conditions 
  are met:

  Redistributions of source code must retain the above copyright notice, 
  this list of conditions and the following disclaimer.

  Neither the name of the copyright holder or other contributors may
  be used to promote products derived from this software without
  specific written prior permission.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES 
  OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND 
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS 
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN 
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN 
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE 
  SOFTWARE.
  
  Contact:   woody@alumni.caltech.edu
  Web page:  http://www.pandawave.com/yaaf/

**********************************************************************/

#include <XConfig.h>
#include <XDraw.h>
#include <XToolTip.h>

/************************************************************************/
/*																		*/
/*	Globals																*/
/*																		*/
/************************************************************************/

XGPopWindow *XGToolTip::gPopWindow;
XGToolTipView *XGToolTip::gPopView;

/************************************************************************/
/*																		*/
/*	Tooltip View														*/
/*																		*/
/************************************************************************/

/*$	XGToolTip::CloseTip
 *
 *		Close the tool tip that I have running
 */

void XGToolTip::CloseTip()
{
	if (gPopWindow) {
		delete gPopWindow;
		gPopWindow = NULL;
	}
}

/*$	XGToolTip::OpenTip
 *
 *		Open this tip with the upper left at the location provided
 */

void XGToolTip::OpenTip(XGView *base, XPoint where, char *what)
{
	XRect r;
	XPoint size;
	XGDraw draw(base);
	draw.SetFont(XGFont::LoadFont(1));
	
	CloseTip();					// close the old tip
	
	gPopView = new XGToolTipView(what);
	size = gPopView->GetPopSize(draw);
	r.left = where.h;
	r.top = where.v;
	r.right = r.left + size.h;
	r.bottom = r.top + size.v;
	
	gPopWindow = new XGPopWindow(base, r, gPopView);
}
